import java.io.*;
import java.net.*;
import java.util.*;
import com.lcmsoft.tftp.*;

/**
 * Simple command line, TFTP server application. This example demonstrates all
 * features of TftpRequest regarding the implementation of TFTP servers,
 * supporting all features of the TFTP standard, specially TFTP options, including
 * how to implement custom options.
 *
 * The server accepts a single, optional argument: the port to listen for requests.
 * The standard TFTP port is used if not provided. Several system properties
 * allows to control other parameters (and the block size option) as well as to
 * activate a custom option "password" that requires a password for any request.
 *
 * Notice that the example just performs error handling regarding inexistent or
 * inaccessible files, sending appropriate error codes to the client. Other runtime
 * errors that may occur are not handled and will abort the server, generating an
 * exception. Likewise, the server assumes a well formed client request.
 *
 * Those omitions are intentional to keep the logic clean and focused to the main
 * points of a server implementation using TftpRequest. A real application would be
 * more robust and would check error conditions, handling any exceptions.
 */
class AdvancedTftpServer {
  
  /* Custom option, password for transferring files. */
  static String PASSWORD = "password";
  
  public static void main( String[] args ) 
  	throws IOException {
	  
    /* Prints a usage message. */
    System.out.println( "Advanced TFTP server example" );
    System.out.println();
    System.out.println( "Usage: AdvancedTftpServer port" );
    System.out.println( "  port - port to listen for TFTP requests (optional)" );
    System.out.println();
    System.out.println( "Other parameters may be specified setting system properties, as follows:" );
    System.out.println( "  tftp.max.block.size - maximum block size accepted" );
    System.out.println( "  tftp.retries - number of retries" );
    System.out.println( "  tftp.timeout.default - default timeout (seconds)" );
    System.out.println( "  tftp.password - password to be passed as a custom option to read/write files" );
    System.out.println( "  tftp.key - registration key" );

    /* Sets the registration key, if provided. */
    Properties properties = System.getProperties();
    String skey = properties.getProperty("tftp.key");
    if ( skey != null )
      TftpRegistrationManager.setRegistrationKey( skey );

    /* If the port parameter was passed on the command line, parse it. */
    int port = (args.length==1)?Integer.parseInt(args[0]):TftpRequest.STANDARD_TFTP_PORT;
	
    mainLoop:
    while ( true ) {
      /* Blocks until a request is received. */
      System.out.println();
      System.out.println( "Waiting for request" );
      TftpRequest request = TftpRequest.getRequest(port);

      /* Sets the number of retries and default timeout, if specified by system properties,
         otherwise left them with their default values. */
      if ( properties.getProperty("tftp.retries") != null )
	request.setRetries( Integer.parseInt(properties.getProperty("tftp.retries")) );
      if ( properties.getProperty("tftp.timeout.default") != null )
	request.setDefaultTimeout( Integer.parseInt(properties.getProperty("tftp.timeout.default")) );
      
      /* Prints the request to the console for information purposes. */
      System.out.println( "Request received from "+request.getHost()+":"+request.getPort() );
      System.out.println( "Operation: "+request.getOperation() );
      System.out.println( "File: "+request.getFileName() );
      System.out.println( "Transfer mode: "+request.getMode() );
      System.out.println( "Received options:" );
      printOptions( request );
      
      /* Process the options received from the client. */
      Iterator iterator = request.optionsNames();
      while ( iterator.hasNext() ) {
	String option = (String)iterator.next();
	if ( option.equals(TftpRequest.TIMEOUT) ) {
	  /* The timeout standard option does not require any custom logic, in this
	     example, so it is just skipped and returned as is to the client. Notice 
	     that it is still necessary to test it otherwise the "catch all" 'else' at
	     the end would delete it. */
	} else if ( option.equals(TftpRequest.TRANSFER_SIZE) ) {
	  /* If the transfer size standard option is set, returns the file size, if it
	     is a read request. If the file does not exist an error will be sent later
	     to the client, so that case does not need to be tested. */
          if ( request.getOperation() == TftpRequest.READ ) {
	    request.setTransferSize( new File(request.getFileName()).length() );
	  }
	} else if ( option.equals(TftpRequest.BLOCK_SIZE) ) {
	  /* If the tftp.max.block.size system property was specified and the requested block
	     size is greater, returns the maximum size as the alternative block size. If the
	     property is not set, does nothing so that the same value is echoed back. */
	  if ( properties.getProperty("tftp.max.block.size") != null ) {
	    int maxBlockSize = Integer.parseInt(properties.getProperty("tftp.max.block.size"));
	    if ( request.getBlockSize() > maxBlockSize )
	      request.setBlockSize( maxBlockSize );
	  }
	} else if ( option.equals(PASSWORD) ) {
	  /* If the tftp.password system propery was specified, checks the password provided
	     by the client, aborting the request if they do not match. If the property is not
	     set, clear the option as it is not supported. */ 
	  if ( properties.getProperty("tftp.password") != null ) {
	    if ( !request.getOptionAsString(PASSWORD).equals(properties.getProperty("tftp.password")) ) {
	      System.out.println( "Wrong password" );
	      request.sendError( TftpErrorException.ERROR_OPTION_NEGOTIATION, "Invalid password" );
	      continue mainLoop;
	    }
	  } else
	    iterator.remove();
	} else
	  iterator.remove();
      }
      /* If the system property tft.password is set but the client did not provide a password,
         aborts the request. */
      if ( (properties.getProperty("tftp.password") != null) &&
           (!request.isOptionSet(PASSWORD)) ) {
	System.out.println( "Password not provided" );
	request.sendError( TftpErrorException.ERROR_OPTION_NEGOTIATION, "Password not provided" );
	continue;
      }

      /* Prints the options to be returned, if any. */
      System.out.println( "Options to return:" );
      printOptions( request );
      
      /* Save the current time to calculate the transfer time later. */ 
      long time = System.currentTimeMillis();
      
      InputStream in;
      OutputStream out;
      if ( request.getOperation() == TftpRequest.READ ) {
	/* If the request is to read a file from the server, tests if the file exists,
	   aborting the request if not. */
	if ( !new File(request.getFileName()).exists() ) {
	  System.out.println( "File not found" );
	  request.sendError( TftpErrorException.ERROR_FILE_NOT_FOUND, "File not found" );
	  continue;
	}
	/* Gets an input stream to read the file from the disk. If an exception occurs,
	   assumes the file is not accessible and sends an "access violation"
	   error to the client. */
	try {
	  in = new BufferedInputStream(new FileInputStream(request.getFileName()));
	} catch ( FileNotFoundException e ) {
	  System.out.println( "File is not accessible" );
	  request.sendError( TftpErrorException.ERROR_ACCESS_VIOLATION, "Access violation" );
	  continue;
	}
	/* Gets an output stream to send the data to the client. */
	out = request.getOutputStream();
      } else {
	/* If the request is to write a file to the server, gets an input stream to
	   receive the file from the client and an output stream to write it to disk.
	   If an exception occurs, assumes the file is not accessible and sends an
	   "access violation" error to the client. */
	in = request.getInputStream();
	try {
	  out = new BufferedOutputStream(new FileOutputStream(request.getFileName()));
	} catch ( FileNotFoundException e ) {
	  System.out.println( "File is not accessible" );
	  request.sendError( TftpErrorException.ERROR_ACCESS_VIOLATION, "Access violation" );
	  continue;
	}
      }

      /* Transfer all data from the input to the output stream, that is, from the
	 disk to the client or vice-versa, depending on the operation (see 
	 above). */
      int b;
      while ( (b = in.read()) != -1 )
	out.write( b );
      in.close();
      out.close();
      
      /* Calculates and prints how much time it took to transfer the file. */
      System.out.println( "Transfer complete" );
      System.out.println( "Elapsed time: "+((System.currentTimeMillis()-time)/1000)+"s" );
    }
  }
  
  
  /**
   * Prints the options of a TftpRequest, passed as parameter, to the console.
   * This method is called before and after the option negotiation so that is
   * possible to visualize the process.
   */
  static void printOptions( TftpRequest request ) {
    Iterator iterator = request.optionsNames();
    while ( iterator.hasNext() ) {
      String option = (String)iterator.next();
      System.out.println("  "+option+": "+request.getOptionAsString(option) );
    }
  }
}