import java.io.*;
import java.net.*;
import java.util.*;
import com.lcmsoft.tftp.*;

/**
 * Simple command line, TFTP server application. This example is the most simple TFTP 
 * server possible. It demonstrates how to handle read and write requests from
 * TFTP clients using just the bare minimum code needed to perform these 
 * operations. For that reason, the server does not support TFTP options and
 * uses the defaults provided by the TftpRequest class.
 *
 * Notice that all error handling code was intentionally removed in order to keep the
 * example short, clean and easy to understand. That also helps to focus the code
 * that does the actual job. Consequently, any error will generate an exception,
 * reported to the console, causing a time-out at the client. A real server would 
 * include code to check error conditions and handle exceptions, reporting them to
 * the client as appropriate.
 */
class BasicTftpServer {
  public static void main( String[] args )
  	throws IOException {

    /* Prints a usage message. */
    System.out.println( "Simple TFTP server example" );
    System.out.println( "The registration key can be set through the system property 'tftp.key'" );
    System.out.println();

    /* Sets the registration key, if provided. */
    Properties properties = System.getProperties();
    String skey = properties.getProperty("tftp.key");
    if ( skey != null )
      TftpRegistrationManager.setRegistrationKey( skey );

    while ( true ) {
      
      /* Blocks until a request is received. */
      System.out.println( "Waiting for request" );
      TftpRequest request = TftpRequest.getRequest();
      
      /* Prints the request to the console for information purposes. */
      System.out.println( "Request received from "+request.getHost()+":"+request.getPort() );
      System.out.println( "Operation: "+request.getOperation() );
      System.out.println( "File: "+request.getFileName() );
      System.out.println( "Transfer mode: "+request.getMode() );
      
      /* Since this example server does not support options, it clear all 
         options of the request, if any, causing the client to use the
	 defaults. */
      request.clearOptions();
      
      /* Save the current time to calculate the transfer time later. */ 
      long time = System.currentTimeMillis();
      
      InputStream in;
      OutputStream out;
      if ( request.getOperation() == TftpRequest.READ ) {
	/* If the request is to read a file from the server, gets an input stream to
	   read the file from the disk and an output stream to send the data to the 
	   client. */
	in = new BufferedInputStream(new FileInputStream(request.getFileName()));
	out = request.getOutputStream();
      } else {
	/* If the request is to write a file to the server, gets an input stream to
	   receive the file from the client and an output stream to write it to disk. */
	in = request.getInputStream();
	out = new BufferedOutputStream(new FileOutputStream(request.getFileName()));
      }

      /* Transfer all data from the input to the output stream, that is, from the
	 disk to the client or vice-versa, depending on the operation (see 
	 above). */
      int b;
      while ( (b = in.read()) != -1 )
	out.write( b );
      in.close();
      out.close();
      
      /* Calculates and prints how much time it took to transfer the file. */
      System.out.println( "Transfer complete" );
      System.out.println( "Elapsed time: "+((System.currentTimeMillis()-time)/1000)+"s" );
      System.out.println();
    }
  }
}