import java.io.*;
import java.net.*;
import java.util.*;
import com.lcmsoft.tftp.*;

/**
 * Command line, TFTP client application. This example demonstrates how to read and
 * write files from/to a TFTP server using the class TftpRequest, as well as all
 * its features, including TFTP options.
 *
 * Notice that all error handling code was intentionally removed in order to keep the
 * example short, clean and easy to understand. That also helps to focus the code
 * that does the actual job. Consequently, any error will generate an exception,
 * reported to the console. A real application would include code to check error
 * conditions, server response and exceptions.
 *
 * The example accepts a command line comprised of three or four arguments:
 *
 *	- the TFTP server address, specified as an IP address or host name
 *	- the operation to perform, either the literal string 'read' or 'write'
 *	- the source file name, including any path information
 *	- the destination file name, including any path information
 *
 * The destination file name can be omitted, in that case the same name provided for
 * the source file will also be used for the destination. Notice that the example
 * does not translate or interpret the file names in any way, so they will be used
 * exactly as provided, including any path information.
 *
 * Other features of the TftpRequest class, as well as TFTP options, can be set
 * through system properties so that it is possible to play with them to see
 * how they work and how they affect the class behaviour. 
 */
class FullTftpClient {
  public static void main( String[] args )
  	throws Exception {
	  
    /* Prints a usage message if the command line is wrong. Notice that only the number
       of parameters are checked. The parameters themselves are not checked and will
       generate an exception if they are not valid. */
    if ( (args.length < 3) || (args.length > 4) ) {
      System.out.println( "Usage: FullTftpClient server op srcfile dstfile" );
      System.out.println( "  server - TFTP server address" );
      System.out.println( "  op - 'read' or 'write'" );
      System.out.println( "  srcfile - source file name" );
      System.out.println( "  dstfile - destination file name (optional)" );
      System.out.println();
      System.out.println( "Other parameters may be specified setting system properties, as follows:" );
      System.out.println( "  tftp.block.size - block size to use" );
      System.out.println( "  tftp.mode - transfer mode" );
      System.out.println( "  tftp.port - remote host port number" );
      System.out.println( "  tftp.retries - number of retries" );
      System.out.println( "  tftp.timeout - timeout (seconds)" );
      System.out.println( "  tftp.timeout.default - default timeout (seconds)" );
      System.out.println();
      System.out.println( "Additionally, custom options may be specified setting properties of the form" );
      System.out.println( "tftp.option.name, replacing 'name' with the name of the option" );
      System.exit( 0 );
    }
    
    /* Extract the parameters from the command line, saving them on appropriate local
       variables. Notice that, regarding the operation parameter, anything different 
       from the string 'read' (case-insensitive) is considered to be 'write'. */
    InetAddress server = InetAddress.getByName(args[0]);
    int operation = args[1].equalsIgnoreCase("read")?TftpRequest.READ:TftpRequest.WRITE;
    String srcFile = args[2];
    String dstFile = args.length>3?args[3]:srcFile;
    
    /* Constructs a TftpRequest object and set the required properties according the
       provided command line parameters. */
    TftpRequest request = new TftpRequest();
    request.setHost( server );
    request.setOperation( operation );
    request.setFileName( operation==TftpRequest.READ?srcFile:dstFile );
    
    /* Sets the transfer size standard option. For read requests, it is always
       set to 0 to instruct the server to return the file size. For write
       requests, it is the size of the file. */
    request.setTransferSize( operation==TftpRequest.READ?0:(new File(srcFile)).length() );
    
    /* Sets additional parameters depending on the system properties. Any parameter
       of the TftpRequest class can be set by a system property. Those that are not
       set this way are left with their default values. */
    Properties properties = System.getProperties();
    if ( properties.getProperty("tftp.block.size") != null )
      request.setBlockSize( Integer.parseInt(properties.getProperty("tftp.block.size")) );
    if ( properties.getProperty("tftp.mode") != null )
      request.setMode( properties.getProperty("tftp.mode") );
    if ( properties.getProperty("tftp.port") != null )
      request.setPort( Integer.parseInt(properties.getProperty("tftp.port")) );
    if ( properties.getProperty("tftp.retries") != null )
      request.setRetries( Integer.parseInt(properties.getProperty("tftp.retries")) );
    if ( properties.getProperty("tftp.timeout") != null )
      request.setTimeout( Integer.parseInt(properties.getProperty("tftp.timeout")) );
    if ( properties.getProperty("tftp.timeout.default") != null )
      request.setDefaultTimeout( Integer.parseInt(properties.getProperty("tftp.timeout.default")) );
    
    /* Sets any custom options specified by system properties (properties 
       of the form 'tftp.option.*'). Notice that the standard options could 
       also be set this way by the properties tftp.option.timeout, 
       tftp.option.blksize and tftp.option.tsize. */
    Enumeration options = properties.propertyNames();
    while ( options.hasMoreElements() ) {
      String option = (String)options.nextElement();
      if ( option.startsWith("tftp.option.") )
	request.setOption( option.substring(12), properties.getProperty(option) );
    }
    
    /* Save the current time to calculate the transfer time later. */ 
    long time = System.currentTimeMillis();

    /* Prints the request, for feedback purposes. */
    System.out.println( "Sending read request to "+request.getHost()+":"+request.getPort() );
    System.out.println( "  file name: "+request.getFileName() );
    System.out.println( "  transfer mode: "+request.getMode() );
    printOptions( request );

    InputStream in;
    OutputStream out;
    if ( operation == TftpRequest.READ ) {
      /* If the request is to read a file from the server, gets an input stream to
         receive the file and an output stream to write it to the disk. */
      in = request.getInputStream();
      out = new BufferedOutputStream(new FileOutputStream(dstFile));
    } else {
      /* If the request is to write a file to the server, gets an input stream to
         read the file from the disk and an output stream to send the data. */
      in = new BufferedInputStream(new FileInputStream(srcFile));
      out = request.getOutputStream();
    }
    
    /* Prints the server answer. */
    System.out.println( "Request accepted" );
    printOptions( request );
    System.out.println( "Request timeout: "+request.getRequestTimeout() );
    System.out.println( "Request block size: "+request.getRequestBlockSize() );

    /* Transfer the file. */
    int b;
    while ( (b = in.read()) != -1 )
      out.write( b );
    in.close();
    out.close();
    
    /* Calculates and prints how much time it took to transfer the file. */
    System.out.println( "Transfer complete" );
    System.out.println( "Elapsed time: "+((System.currentTimeMillis()-time)/1000)+"s" );
  }
  
  
  /**
   * Prints the options of a TftpRequest, passed as parameter, to the console.
   * This method is called before and after the option negotiation so that is
   * possible to visualize the process.
   */
  static void printOptions( TftpRequest request ) {
    Iterator iterator = request.optionsNames();
    if ( !iterator.hasNext() ) {
      System.out.println( "The server returned no options" );
      return;
    }
    while ( iterator.hasNext() ) {
      String option = (String)iterator.next();
      System.out.println("  "+option+": "+request.getOptionAsString(option) );
    }
  }
}
